/* Copyright (c) 2000-2009 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

(function() {

    /**
     * A href|src file handler - to help in opening files. Examples:
     *    href="#main"
     *    src="foobar.css"
     *    href="script.js"
     */

    var _bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
        .getService(Components.interfaces.nsIStringBundleService)
        .createBundle("chrome://komodo/locale/hyperlinks/hyperlinks.properties");

    /**
     * Open the href location.
     */
    function filename_jump_handler(filepath) {
        if (typeof(filepath) == 'undefined') {
            var match = src_href_handler.regex_match;
            filepath = match[2];
        }
        if (filepath) {
            /**
             * @type {Components.interfaces.koIScintillaView}
             */
            var view = ko.views.manager.currentView;
            if (filepath[0] == "#") {
                // It's within this document (an anchor).
                // #1 - Look for an id=""
                var sm = view.scimoz;
                var text = sm.text;
                var pos = text.search(new RegExp("\\sid\\s?=\\s?['\"]" + filepath.substr(1) + "['\"][\\s>]", "i"));
                if (pos >= 0) {
                    var scimoz_pos = ko.stringutils.bytelength(text.substr(0, pos));
                    sm.gotoPos(scimoz_pos);
                    return;
                }
                // #2 - Look for a '<a name=...'
                var pos = text.search(new RegExp("<a\\s+(.*?\\s)?name\\s?=\\s?['\"]" + filepath.substr(1) + "['\"][\\s>]", "i"));
                if (pos >= 0) {
                    var scimoz_pos = ko.stringutils.bytelength(text.substr(0, pos));
                    sm.gotoPos(scimoz_pos);
                    return;
                } else {
                    ko.statusBar.AddMessage(_bundle.formatStringFromName("noAnchorFound.message", [filepath], 1),
                                            "hyperlinks", 3000, true);
                }
            } else if (filepath[0] == "\\" || filepath[0] == "/") {
                // It's an absolute path.
                ko.open.URI(filepath);
            } else {
                // It's a relative path.
                // #1 - Check current directory.
                var osPathSvc = Components.classes["@activestate.com/koOsPath;1"].getService(Components.interfaces.koIOsPath);
                var view_filepath = view.document.file.path;
                var try_filepath = osPathSvc.join(ko.uriparse.dirName(view_filepath), filepath);
                if (osPathSvc.exists(try_filepath)) {
                    ko.open.URI(try_filepath);
                    return;
                }
                // #2 TODO: check against a list of known places (images, css,
                //          etc...).

                // For now we just try to open the file - perhaps using a mapped
                // URI will get them to the right file.
                ko.open.URI(filepath);
            }
        }
    }
    function src_href_jump_fn() {
        var match = src_href_handler.regex_match;
        var filepath = match[2];
        filename_jump_handler(filepath);
    }
    var src_href_handler = new ko.hyperlinks.RegexHandler(
                            "Src and Href handler",
                            new RegExp("\\b(src|href)=[\\\"'](.*?)[\\\"']", "i"),
                            src_href_jump_fn,
                            null,  /* Use the found string instead of a replacement. */
                            null,  /* All language types */
                            Components.interfaces.ISciMoz.INDIC_PLAIN,
                            RGB(0x60,0x90,0xff));
    ko.hyperlinks.addHandler(src_href_handler);


    /**
     * A generic file handler - to help in opening files. Examples:
     *    file:///foo/bar.txt
     *    /foo/bar.txt
     *    C:\foo\bar.txt
     *    \\foo\bar.txt
     */
    ko.hyperlinks.addHandler(
        new ko.hyperlinks.RegexHandler(
            "File handler",
            new RegExp("(file:|/|\\[A-Z]:\\\\|\\\\)[^'\"<>()[\\]\\s]+", "i"),
            ko.open.URI,
            null,  /* Use the found string instead of a replacement. */
            null,  /* All language types */
            Components.interfaces.ISciMoz.INDIC_PLAIN,
            RGB(0x60,0x90,0xff))
    );


    /**
     * A PHP include file handler - to help in opening files. Examples:
     *    include('functions/myfile.php');
     *    require_once "foo/bar/myfile.php";
     */
    function php_jump_handler() {
        var match = php_include_handler.regex_match;
        var filepath = match[3];
        filename_jump_handler(filepath);
    }
    var php_include_handler = new ko.hyperlinks.RegexHandler(
            "File handler",
            new RegExp("(include|require|include_once|require_once)\\s*(\\(\\s*)?[\"'](.*?)[\"']\\s*\\)?", "i"),
            php_jump_handler,
            null,  /* Use the found string instead of a replacement. */
            null,  /* All language types */
            Components.interfaces.ISciMoz.INDIC_PLAIN,
            RGB(0x60,0x90,0xff));
    ko.hyperlinks.addHandler(php_include_handler);

})();
