/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

// A project extension to manage source code control operations in Komodo.

if (typeof(ko)=='undefined') {
    var ko = {};
}
if (typeof(ko.projects)=='undefined') {
    ko.projects = {};
}

(function() {

var log = ko.logging.getLogger('peSCC');
var _bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
      .getService(Components.interfaces.nsIStringBundleService)
      .createBundle("chrome://komodo/locale/project/peSCC.properties");
var _prefs = Components.classes["@activestate.com/koPrefService;1"].
                    getService(Components.interfaces.koIPrefService).prefs;
var _obSvc = Components.classes["@mozilla.org/observer-service;1"].
             getService(Components.interfaces.nsIObserverService);

//---- SCC project extension implementation.

function peSCC() {
    this.name = 'SCC';
    //log.setLevel(ko.logging.LOG_DEBUG);
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
peSCC.prototype.constructor = peSCC;

peSCC.prototype.init = function() {
    // register our command handlers
    this.registerCommands();

    // this adds columns to the project/toolbox tree's
    var em = ko.projects.extensionManager;
    em.setDatapoint(_bundle.GetStringFromName("sccStatusTreeColLabel"), 'sccStatus');
    em.setDatapoint(_bundle.GetStringFromName("sccLocalRevisionTreeColLabel"), 'sccLocalRevision');
    em.setDatapoint(_bundle.GetStringFromName("sccDepotRevisionTreeColLabel"), 'sccDepotRevision');
    em.setDatapoint(_bundle.GetStringFromName("sccActionTreeColLabel"), 'sccAction');
}

peSCC.prototype.registerCommands = function() {
    var em = ko.projects.extensionManager;
    em.registerCommand('cmd_SCCedit', this);
    em.registerCommand('cmd_SCCadd', this);
    em.registerCommand('cmd_SCCremove', this);
    em.registerCommand('cmd_SCCupdate', this);
    em.registerCommand('cmd_SCCcommit', this);
    em.registerCommand('cmd_SCCdiff', this);
    em.registerCommand('cmd_SCChistory', this);
    em.registerCommand('cmd_SCCrevert', this);
    em.registerCommand('cmd_SCCcheckout', this);

    em.registerCommand('cmd_SCCadd_folder', this);
    em.registerCommand('cmd_SCCupdate_folder', this);
    em.registerCommand('cmd_SCCcommit_folder', this);
    em.registerCommand('cmd_SCCdiff_folder', this);
    em.registerCommand('cmd_SCCrevert_folder', this);
    em.registerCommand('cmd_SCCcheckout_folder', this);
}

peSCC.prototype.registerEventHandlers = function() {
}

peSCC.prototype.registerMenus = function() {
    var menupopup = document.getElementById('menu_SCCmenu_context');
    if (!menupopup) {
        log.error("Couldn't get 'menu_SCCmenu'");
        return;
    }
    var folderpopup = document.getElementById('menu_SCCmenu_folder_context');
    if (!folderpopup) {
        log.error("Couldn't get 'menu_SCCmenu_folder'");
        return;
    }
    var em = ko.projects.extensionManager;
    em.addMenuItem(Components.interfaces.koIPart_SCC_file, document.createElement('menuseparator'));
    em.addMenuItem(Components.interfaces.koIPart_SCC_file, menupopup);

    // grab the child menuitems individually
    var popupChildren = folderpopup.firstChild.firstChild;
    // XXX HACK to avoid the "Add" command on a container, we check the
    //     "nextSibling" in the while loop
    while (popupChildren.nextSibling) {
        em.addMenuItem(Components.interfaces.koIPart_changelist, popupChildren);
        popupChildren = popupChildren.nextSibling;
    }

    em.addMenuItem(Components.interfaces.koIPart_SCC_directory, document.createElement('menuseparator'));
    em.addMenuItem(Components.interfaces.koIPart_SCC_directory, folderpopup);
}

/* This takes an item, and determines what type of scc
   flags we are dealing with. returns true if we can do anything, or
   false if we cannot do anything */
peSCC.prototype.determineItemSCCSupport = function(/* koIPart */item, sccObj) {
    // prevent non-cvs directories from having to go through this torture
    var file = item.getFile();
    if (file && file.sccExclude) {
        return false;
    }
    // assume the worst =)
    sccObj.isscc_file = false;
    sccObj.isscc_directory = false;
    sccObj.isscc_container = false;
    sccObj.isscc_collection = false;

    var ok;
    try {
        ok = item.QueryInterface(Components.interfaces.koIPart_SCC_file) != null;
        if (ok) {
            sccObj.isscc_file = true;
        }
    } catch(e) {}
    try {
        ok = item.QueryInterface(Components.interfaces.koIPart_SCC_collection) != null;
        if (ok) {
            sccObj.isscc_collection = true;
        }
    } catch(e) {}
    try {
        ok = item.QueryInterface(Components.interfaces.koIPart_SCC_directory) != null;
        if (ok) {
            sccObj.isscc_directory = true;
        }
    } catch(e) {}
    sccObj.isscc_container = sccObj.isscc_directory || sccObj.isscc_collection;
    sccObj.isscc_file = sccObj.isscc_file || !sccObj.isscc_container;
    sccObj.has_scc = file && file.sccType != '';
    sccObj.is_scc = sccObj.isscc_file || sccObj.isscc_container;
    if (!sccObj.is_scc) {
        return false;
    }

    sccObj.can_scc = sccObj.is_scc || sccObj.has_scc || (file && file.sccDirType != '');
    sccObj.missing = file && !file.exists;

    var in_repository = null;
    if (item.url && !sccObj.can_scc && !sccObj.missing) {
        sccObj.can_scc = sccObj.has_scc || (file && file.sccDirType);
    }

    if (file && file.sccExclude) return false;
    return true;
}

peSCC.prototype.supportsCommand = function(command, item) {
    // do we have either cvs or scc?

    var enabled_scc_components = ko.scc.getAvailableSCCComponents(true);
    if (!enabled_scc_components.length) {
        // No scc components are enabled for functional.
        return false;
    }

    if (command == 'cmd_SCCcheckout') {
        return true;
    }

    // If scintilla has focus, should be talking about the file in scintilla
    if (!item) {
        if (ko.projects.getFocusedProjectView()) {
            item = ko.projects.active.getSelectedItem();
            if (!item) {
                return false;
            }
        } else {
            var view = ko.views.manager.currentView;
            if (!view || !view.document || view.document.isUntitled) {
                // This is an unsaved file -- no SCC commands can apply.
                return false;
            }
            item = view.item;
            item.url = view.document.file.URI;
        }
    }
    // XXX look into cacheing sccObj into the item
    var sccObj = new Object();
    if (!this.determineItemSCCSupport(item,sccObj)) {
        return false;
    }
    var isfoldercmd = command.indexOf('_folder')>1;
    /* if cvs, and this is a directory, then the directory MUST be in cvs
    if (isfoldercmd) {
      dump("URL: "+item.url+"\n");
      ko.logging.dumpObject(sccObj);
      dump("  isfoldercmd "+isfoldercmd+"\n");
      dump("  folders_in_repository "+folders_in_repository+"\n");
    }
    */
    /* folder handling is different between p4 and cvs */
    var file = item.getFile();
    if (isfoldercmd && !sccObj.isscc_container) {
        return false;
    } else if (!isfoldercmd && !file) {
        return false;
    }
    var folders_in_repository = file && (file.sccType == 'cvs' || file.sccDirType == 'cvs' ||
                        file.sccType == 'svn' || file.sccDirType == 'svn');
    var folders_revertable = sccObj.isscc_collection || file && (file.sccType == 'p4' || file.sccDirType == 'p4');

    var foldercmd_ok = (sccObj.isscc_container && isfoldercmd) &&
            ((folders_in_repository && (sccObj.isscc_directory || sccObj.isscc_collection)) ||
            (!folders_in_repository && sccObj.can_scc));
    var retval;
    switch (command) {
    case 'cmd_SCCedit':
        retval = sccObj.has_scc &&
            (file.sccAction == '' || file.sccAction == 'integrate')
            && (file.sccType == 'p4' ||
                file.sccType == 'cvs' && file.isReadOnly)
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCadd_folder':
        retval = sccObj.can_scc && !sccObj.has_scc && isfoldercmd
                && sccObj.isscc_directory && folders_in_repository;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCadd':
        retval = sccObj.can_scc && !sccObj.has_scc;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCupdate_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCupdate':
        retval = sccObj.has_scc;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCremove':
        return sccObj.has_scc && !file.sccNeedSync &&
            file.sccAction == '';
        break;
    case 'cmd_SCCdiff_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCChistory':
        retval = sccObj.has_scc;
        return retval;
        break;
    case 'cmd_SCCdiff':
        retval = sccObj.has_scc &&(file.sccNeedSync ||
                file.sccAction == 'edit' || file.sccAction == 'integrate' ||
                file.sccAction == 'conflict');
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCrevert_folder':
        return foldercmd_ok && folders_revertable;
        break;
    case 'cmd_SCCrevert':
        retval = sccObj.has_scc && (file.sccNeedSync ||
                file.sccAction != '');
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCcommit_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCcommit':
        return sccObj.isscc_container || (sccObj.has_scc && file.sccAction != ''
                                          && (!file.sccNeedSync ||
                                              file.sccAction == 'delete'));
        break;
    case 'cmd_SCCcheckout_folder':
        return true;
    default:
        break;
    }
    return false;
}

peSCC.prototype.isCommandEnabled = peSCC.prototype.supportsCommand;

/* recursivly build a list of items that are being used in
   this scc operation */
peSCC.prototype._addVirtualFolderChildren = function(item)
{
    var items = [];
    try {
        var children = new Array();
        item.getChildren(children, new Object());
        children = children.value;
    } catch(e) {
        return items;
    }
    if (children.length < 1) return items;

    var child;
    for (var i = 0; i < children.length; i++) {
        child = children[i];
        if (child.url) {
            items.push(child);
        } else {
            var c = this._addVirtualFolderChildren(child);
            if (c.length > 0) {
                items = items.concat(c);
            }
        }
    }
    return items;
}

peSCC.prototype.doCommand = function(command) {
    if (command.substr(0, 15) == 'cmd_SCCcheckout') {
        window.setTimeout(function (self, command) { self._doSccCheckout(command); }, 1, this, command);
    } else {
        window.setTimeout(function (self, command) { self._doCommand(command); }, 1, this, command);
    }
}

function _perform_scc_command_on_urls(command, sccType, collection, urls,
                                      localFileVersions, callback) {

    log.debug('sccType: '+sccType +', urls:\n  ' + urls.join('\n  '));

        // Get the koISCC service object
    var cid = "@activestate.com/koSCC?type=" + sccType + ";1";
    var sccSvc = Components.classes[cid].getService(Components.interfaces.koISCC);
    if (!sccSvc || !sccSvc.isFunctional) {
        return;
    }
    var isfoldercmd = command.indexOf('_folder')>1;
    try {
        switch (command) {
        case 'cmd_SCCedit':
            ko.scc.Edit(sccSvc, urls, callback);
            break;
        case 'cmd_SCCadd_folder':
            if (!isfoldercmd) return;
            // else fall through
        case 'cmd_SCCadd':
            ko.scc.Add(sccSvc, urls, callback);
            break;
        case 'cmd_SCCremove':
            ko.scc.Remove(sccSvc, urls, callback);
            break;
        case 'cmd_SCCrevert_folder':
            if (!isfoldercmd) return;
            // else fall through
        case 'cmd_SCCrevert':
            ko.scc.Revert(sccSvc, urls, callback);
            break;
        case 'cmd_SCCupdate_folder':
            if (!isfoldercmd) return;
            // else fall through
        case 'cmd_SCCupdate':
            ko.scc.Update(sccSvc, urls, callback);
            break;
        case 'cmd_SCCdiff_folder':
            if (!isfoldercmd) return;
            // else fall through
        case 'cmd_SCCdiff':
            ko.scc.Diff(sccSvc, urls, callback);
            break;
        case 'cmd_SCCcommit_folder':
            if (!isfoldercmd) return;
            // else fall through
        case 'cmd_SCCcommit':
            if (!callback && collection != null) {
                callback = function (result, data) {
                    if (result != Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) return;
                    // remove scc urls now
                    for (i=0; i < urls.length; i++) {
                        var c = collection.getChildByAttributeValue('url', urls[i], true);
                        if (c) {
                            c.parent.removeChild(c);
                        }
                    }
                    var collection_parent = collection.parent;
                    if (collection.isEmpty()) {
                        collection_parent.removeChild(collection);
                    }
                    ko.toolboxes.scc.viewMgr.view.refresh(collection_parent);
                }
            }
            var commit_message = '';
            if (collection && collection.prefset.hasStringPref("changelist_description")) {
                commit_message = collection.prefset.getStringPref("changelist_description");
            }
            ko.scc.Commit(sccSvc, urls, commit_message, callback);
            break;
        case 'cmd_SCChistory':
            ko.scc.History(sccSvc, urls, localFileVersions);
            break;
        default:
            break;
        }
    } catch (e) {
        alert(_bundle.formatStringFromName("sccError.alert", [e], 1));
        log.error(e);
    }
}

function _scc_add_parent_directories(urls, depth, callback) {
    // Check if the parent folder of the targeted item(s) are under
    // SCC, we can then prompt to add the directory first and then
    // add the targeted item(s) afterwards.
    log.debug("_scc_add_parent_directories: depth: " + depth + ", urls: " + urls);
    var fileSvc = Components.classes["@activestate.com/koFileService;1"].
                        getService(Components.interfaces.koIFileService);
    var dir_url;
    var parent_dir_url;
    var dir_urls = [];
    var koFile;
    var dir_koFile;
    var parent_dir_koFile;
    var dir_koFiles = [];
    var unknownSccType = false;
    var sccType = "";
    for (var i=0; i < urls.length; i++) {
        koFile = fileSvc.getFileFromURI(urls[i]);
        dir_url = ko.uriparse.dirName(urls[i]);
        dir_urls.push(dir_url);
        dir_koFile = fileSvc.getFileFromURI(dir_url);
        parent_dir_url = ko.uriparse.dirName(dir_url);
        parent_dir_koFile = fileSvc.getFileFromURI(parent_dir_url);
        if (sccType && parent_dir_koFile.sccDirType && (parent_dir_koFile.sccDirType != sccType)) {
            ko.statusBar.AddMessage("Mixed SCC types are not supported: '" + parent_dir_koFile.sccDirType + "', and '" + sccType + "'",
                                    "scc", 5000, true, false);
            return;
        } else if (!sccType && parent_dir_koFile.sccDirType) {
            sccType = parent_dir_koFile.sccDirType;
        } else {
            // There is no f.sccDirType on this file.
            unknownSccType = true;
        }
        if (dir_koFiles.indexOf(dir_koFile) < 0) {
            dir_koFiles.push(dir_koFile);
            //dump("dir_koFile: " + dir_koFile.path + "\n");
        }
    }
    log.debug("_scc_add_parent_directories: depth: " + depth +
              ", sccType: " + sccType);
    if (!unknownSccType) {
        // Perform the scc add directory operations then. We know which SCC
        // handler we need to use.
        var add_dir_callback = function (result, data) {
            if (result != Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                callback(result, null, "Failed: could not add parent directory");
                return;
            }
            callback(result, sccType, null);
        }
        _perform_scc_command_on_urls("cmd_SCCadd_folder", sccType, null,
                                     dir_urls, null, add_dir_callback);
        return;
    }

    // Need to find out if the parent directory is under SCC control. Launch
    // the asynchronous update then, which will call back later to check.
    var fileStatusSvc = Components.classes["@activestate.com/koFileStatusService;1"].
                        getService(Components.interfaces.koIFileStatusService);
    var filesvc_callback = (function(koFiles, parent_urls) {
        log.debug("_scc_add_parent_directories: depth: " + depth +
                  ", performing file status check");
        return {
            "notifyDone": function() {
                log.debug("_scc_add_parent_directories: depth: " + depth +
                          ", notifyDone");
                // The scc status has been checked.
                var scc_type;
                var f;
                for (var i=0; i < koFiles.length; i++) {
                    f = koFiles[i];
                    if (!f.sccDirType) {
                        log.debug("_scc_add_parent_directories: depth: " +
                                  depth + ", parent is not under scc");
                        if (depth < 3) {
                            var cb = function(result, scc_type, message) {
                                log.debug("_scc_add_parent_directories: depth: " +
                                          (depth+1) + ", cb: scc_type: " + scc_type);
                                var add_dir_callback = function (result, data) {
                                    if (result != Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                                        callback(result, null, "Failed: could not add parent directory");
                                        return;
                                    }
                                    callback(result, scc_type, null);
                                }
                                // Call the scc handler to do the work on the given urls.
                                if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                                    _perform_scc_command_on_urls("cmd_SCCadd_folder", scc_type, null, parent_urls, null, add_dir_callback);
                                } else {
                                    callback(Components.interfaces.koIAsyncCallback.RESULT_ERROR,
                                             null,
                                             "Failed: No parent directory is under SCC control");
                                }
                            }
                            _scc_add_parent_directories(parent_urls, depth+1, cb);
                        } else {
                            callback(Components.interfaces.koIAsyncCallback.RESULT_ERROR,
                                     null,
                                     "Failed: No parent directory is under SCC control");
                        }
                        return;
                    } else if (scc_type && (f.sccDirType != scc_type)) {
                        log.debug("_scc_add_parent_directories: depth: " +
                                  depth + ", mixed types found: " + scc_type);
                        callback(Components.interfaces.koIAsyncCallback.RESULT_ERROR,
                                 null,
                                 "Failed: Mixed SCC types found: '" +
                                 f.sccDirType + "', and '" + scc_type + "'");
                        return;
                    } else if (!scc_type && f.sccDirType) {
                        scc_type = f.sccDirType;
                    }
                }
                callback(Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL,
                         scc_type, null);
            }
        }
    })(dir_koFiles, dir_urls);
    ko.statusBar.AddMessage("Target directory not under SCC, checking the parent...",
                            "scc", 5000, true, false);
    ko.scc.logger.append("Checking SCC status of parents:\n  " + dir_urls.join("\n  ") + "\n");
    fileStatusSvc.updateStatusForFiles(dir_koFiles.length,
                                       dir_koFiles,
                                       false, filesvc_callback);
}

peSCC.prototype._doCommand = function(command, scc_items /* null */) {
    var cmd = null;
    var error = null;
    var urls=[];
    var _sccUrls = {};
    var _sccLocalVersions = {}; /* Only used for scc history */
    var _allURLS = [];
    var file;
    var item = null;
    var items = [];
    var i;
    var sccType;

    log.debug("doCommand "+command);

    // Iterate over all the targeted items (project items, files, ...) and
    // gather the urls we are going to perform SCC actions upon

    // XXX FIXME this is an issue since projects do not necessarily have focus
    // on a context menu operation
    if (!scc_items) {
        if (ko.projects.getFocusedProjectView()) {
            // Get the items from the focused project
            items = ko.projects.active.getSelectedItems();
        } else {
            // Get the items from the currently focused file
            //dump("Scintilla has focus -- actions should be based on the current file\n");
            var view = ko.views.manager.currentView;
            if (view && !view.document.isUntitled) {
                item = ko.projects.manager.findItemByURL(view.document.file.URI);
                if (!item) {
                    item = view.item;
                    item.url = view.document.file.URI;
                    item.project = null;
                }
                items.push(item);
            }
        }
    
        if (items.length < 1) return;

    } else {
        items = scc_items;
    }


    var collection = null;
    for (i=0; i<items.length; i++) {
        item = items[i];
        // container support
        // check for scc containers, add the children of the container
        // to our items array.
        try {
            collection = item.QueryInterface(Components.interfaces.koIPart_SCC_collection);
        } catch(e) {}
        if (collection != null) {
            var c = this._addVirtualFolderChildren(item);
            if (c.length > 0) {
                items = items.concat(c);
            }
        }
    }

    // We have items, now work out the urls for the scc handlers
    var isfoldercmd = command.indexOf('_folder')>1;
    var found_supported_item = false;
    for (i=0; i<items.length; i++) {
        item = items[i];
        file = item.getFile();
        sccType = file && (file.sccType || file.sccDirType);
        if (!sccType || file.sccExclude) {
            log.debug("doCommand unable to determineItemSCCSupport "+item.url);
            ko.scc.logger.append("No SCC handler found for: " + item.url + "\n");
            continue
        }

        found_supported_item = true;

        if (typeof(_sccUrls[sccType]) == 'undefined') {
            _sccUrls[sccType] = [];
            _sccLocalVersions[sccType] = [];
        }
        if (isfoldercmd && item.type == 'project') {
            var url = ko.uriparse.dirName(item.url)
            _sccUrls[sccType].push(url);
            _sccLocalVersions[sccType].push(""); /* No version, it's a collection */
            _allURLS.push(url);
            // add the dialogs children
            //items = items.concat(this._addVirtualFolderChildren(item));
        } else {
            _sccUrls[sccType].push(item.url);
            _sccLocalVersions[sccType].push(file.sccLocalRevision);
            _allURLS.push(item.url);
        }
    }

    if (!found_supported_item) {
        if (command == "cmd_SCCadd" || command == "cmd_SCCadd_folder") {
            ko.statusBar.AddMessage("Checking SCC status of the parent folder...",
                                    "scc", 15000, true, false);
            var urls = [item.url for each (item in items)];
            var callback = function(result, scc_type, message) {
                // Call the scc handler to do the work on the given urls.
                if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                    log.debug("Added parent " + scc_type + " directories to " +
                              "scc, now adding: " + urls);
                    var sccurls = _sccUrls[scc_type];
                    _perform_scc_command_on_urls(command, scc_type, null, urls);
                    ko.statusBar.AddMessage("Added parent folders to " + scc_type,
                                            "scc", 5000, true, false);
                } else {
                    ko.scc.logger.append(message + "\n");
                    ko.statusBar.AddMessage(message, "scc", 5000, true, false);
                }
            }
            // The below setTimeout used in order to allow the above statusbar
            // message to be displayed.
            window.setTimeout(function(urls_, callback_) {
                    _scc_add_parent_directories(urls_, 0, callback_);
                }, 1, urls, callback);
        } else {
            ko.statusBar.AddMessage("No supported SCC items found, ignoring the command.",
                                    "scc", 5000, true, false);
        }
        return;
    }

    if ((command != 'cmd_SCCrevert') &&
        (command != 'cmd_SCChistory') &&
	!ko.views.manager.offerToSave(_allURLS, /* urls */
                      'Save Modified Files?', /* title */
                      'Save Modified Files Before Source Code Control Action', /* prompt */
                      "save_on_scc", /* pref */
                      false, /* skip projects */
                      false /* about to close */
                      )) {
        return;
    }
    
    // We've found all the SCC urls, now perform the scc actions for these urls.

    var output = '';
    var title;
    var sccType;
    var sccSvc;

    log.debug('selected for scc operation '+_sccUrls);
    for (sccType in _sccUrls) {
        urls = _sccUrls[sccType];
        if (urls.length <= 0)
            continue;
        // Call the scc handler to do the work on the given urls.
        _perform_scc_command_on_urls(command, sccType, collection, urls, _sccLocalVersions[sccType]);
    }
    // Ensure the cached information gets updated, so the SCC toolbars and menu
    // are properly set.
    // Fixes bug: http://bugs.activestate.com/show_bug.cgi?id=48417
    ko.views.manager.resetLastViewCache();

    window.updateCommands('SCC'); // We've changed things hopefully, and no one else knows.
}

peSCC.prototype._doSccCheckout = function(command) {
    var checkoutLocation = null;
    if (command == 'cmd_SCCcheckout_folder') {
        items = ko.projects.active.getSelectedItems();
        if (items.length <= 0) {
            return;
        }
        if (items.length != 1) {
            ko.dialogs.alert(_bundle.GetStringFromName("sccCheckoutMultipleFoldersError.alert"));
            return;
        }
        checkoutLocation = items[0].url;
    }
    ko.scc.Checkout(checkoutLocation);
}

// This object is now used by the places extension.
this.SCC = new peSCC();
ko.projects.registerExtension(this.SCC);

}).apply(ko.projects);
